/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2008 Apr 05\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: job_local.c,v 1.3 2008/04/21 14:43:18 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <mife.h>
#include <err.h>
#include <multilar.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"
#include "pgob.h"

static int
local_none(pgoblin_exenv *exenv, pgoblin_r *r) {
    return(EX_OK);
}

static u_int32_t
local_test(pgoblin_command cmd) {
    u_int32_t e;

    if  (cmd == PGOBLIN_fork) e = PGOBLIN_TVFORK; else e = PGOBLIN_TOFORK | PGOBLIN_TOEXIT;
    return(e);
}

static int
local_exec(pgoblin_exenv *exenv, pgoblin_r *r, char **argv) {
    int ex = 0;

    ifBLIN_QV5(exenv->options->flags) fprintf( stderr, "+local_exec %d %d %d %d %08X argv %s\n"
                                      , r[PGO_COUT].r, r[PGO_CCTL].r, r[PGO_CIN].r, r[PGO_CJOB].r
                                      , exenv->exeflags, argv ? "OK" : "NULL"
                                      );
    ifBLIN_QV6(exenv->options->flags) {
        int z;
        for (z = 0; argv[z]; z++) fprintf(stderr, "%d=%s\n", z, argv[z]);
    }
    if  (argv && !ex) {
        int g;
        /************************************************
         * local_exec  ,    *
         ************************************************/
    ;   if  (dup2(IO_IN->mife->fd, fileno(stdin)) < 0) {
    ;       ifBLIN_QV1(exenv->options->flags) warn("dup2 #4");
    ;       ex = -EX_OSERR;
    ;       exit(ex);
    ;   }
    ;   if  (dup2(IO_OUT->onu, fileno(stdout)) < 0) {
    ;       ifBLIN_QV1(exenv->options->flags) warn("dup2 #3");
    ;       ex = -EX_OSERR;
    ;       exit(ex);
    ;   }
    ;   if  (exenv->options->flags & PGOBLIN_FORK_WAIT) {
    ;       ifBLIN_QV3(exenv->options->flags)
                fprintf(stderr, "sleep %us in fork\n", exenv->options->flags & PGOBLIN_FORK_WAIT);
    ;       sleep(exenv->options->flags & PGOBLIN_FORK_WAIT);
    ;   }
    ;   for (g = 0; g < PGOBLIN_REGSIZE; g++) {
#           define R exenv->options->io[g]
    ;       if  (R.mife && (R.mife->fd != fileno(stdin)) && (R.mife->fd != fileno(stdout))) {
    ;           if  ((R.flags & PGOBLIN_FREEMIFE) && (ex = mife_close(R.mife)))
                    ifBLIN_QV1(exenv->options->flags) warn("mife_close in exec");
    ;       }
    ;       if  (R.clo && (R.onu != fileno(stdin)) && (R.onu != fileno(stdout))) {
    ;           if  ((R.flags & PGOBLIN_FREEOUTS) && (ex = R.clo(R.onu)))
                    ifBLIN_QV1(exenv->options->flags) warn("clo in exec");
    ;       }
#           undef R
    ;   }
    ;   execve(argv[0], argv, exenv->options->envp);
    ;   ex = EX_OSERR;
        ifBLIN_QV1(exenv->options->flags) warn("execve %s", argv[0]);
        exit(ex);
    }
    ifBLIN_QV5(exenv->options->flags) fprintf( stderr, "-local_exec %d\n", ex);
    return((ex > 0) ? 0 : ex);
}

static int
local_wait(pgoblin_exenv *exenv, pgoblin_r *r) {
    int ex = EX_OK;
    pgoblin_jobreg *rj;
    pgoblin_job *j;
    size_t i;

    rj = R_JOB;
    if  (rj) {
        ifBLIN_QV5(rj->flags) fprintf(stderr, "+local_wait\n");
        for (i = 0; i <  MULAR_NEXT(rj->pids); i++) {
            j = mular_getix(rj->pids, i);
            if  (!(j->pid)) {
                ifBLIN_QV1(rj->flags) warnx("No pid in %d", j->cmdn);
                errno = EINVAL;
                ex = EX_SOFTWARE;
            } else if ((j->cmdn >= 0) && (j->status == PGOBLIN_NOSTATUS)) {
                if  ((ex = waitpid(j->pid, &(j->status), 0)) < 0) {
                    ifBLIN_QV1(rj->flags) warn("Pid %d in %d", j->pid, j->cmdn);
                    ex = EX_OSERR;
                } else if (ex == j->pid) {
                    ex = EX_OK;
                } else {
                    ifBLIN_QV1(rj->flags) warn("Pid %d<>%d in %d", ex, j->pid, j->cmdn);
                    ex = EX_OSERR;
        }   }   }
        ifBLIN_QV5(rj->flags) fprintf(stderr, "-pgoblin_wait %d\n", ex);
    }
    return(ex);
}

pgoblin_jobs pgoblin_job_local =
{ "local"
, local_test
, local_none
, local_exec
, local_wait
, local_none
};
